@@ -11,13 +11,26 @@ from iminuit.cost import (
     LeastSquares,
     Constant,
     NormalConstraint,
-    _log_poisson_part,
+    BarlowBeestonLite,
+    multinominal_chi2,
+    barlow_beeston_lite_chi2_jsc,
+    _soft_l1_loss,
     PerformanceWarning,
 )
 from collections.abc import Sequence
 
 stats = pytest.importorskip("scipy.stats")
 norm = stats.norm
+truncexpon = stats.truncexpon
+
+
+def mvnorm(mux, muy, sx, sy, rho):
+    C = np.empty((2, 2))
+    C[0, 0] = sx**2
+    C[0, 1] = C[1, 0] = sx * sy * rho
+    C[1, 1] = sy**2
+    m = [mux, muy]
+    return stats.multivariate_normal(m, C)
 
 
 def expon_cdf(x, a):
@@ -79,6 +92,23 @@ def test_UnbinnedNLL(unbinned, verbose, model):
     assert_equal(m.fmin.reduced_chi2, np.nan)
 
 
+def test_UnbinnedNLL_2D():
+    def model(x_y, mux, muy, sx, sy, rho):
+        return mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)
+
+    truth = 0.1, 0.2, 0.3, 0.4, 0.5
+    x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T
+
+    cost = UnbinnedNLL((x, y), model)
+    m = Minuit(cost, *truth)
+    m.limits["sx", "sy"] = (0, None)
+    m.limits["rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+
+    assert_allclose(m.values, truth, atol=0.02)
+
+
 @pytest.mark.parametrize("verbose", (0, 1))
 @pytest.mark.parametrize("model", (logpdf, pdf))
 def test_ExtendedUnbinnedNLL(unbinned, verbose, model):
@@ -104,6 +134,23 @@ def test_ExtendedUnbinnedNLL(unbinned, verbose, model):
     assert_equal(m.fmin.reduced_chi2, np.nan)
 
 
+def test_ExtendedUnbinnedNLL_2D():
+    def model(x_y, n, mux, muy, sx, sy, rho):
+        return n * 1000, n * 1000 * mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)
+
+    truth = 1.0, 0.1, 0.2, 0.3, 0.4, 0.5
+    x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 1000)).T
+
+    cost = ExtendedUnbinnedNLL((x, y), model)
+    m = Minuit(cost, *truth)
+    m.limits["n", "sx", "sy"] = (0, None)
+    m.limits["rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+
+    assert_allclose(m.values, truth, atol=0.1)
+
+
 @pytest.mark.parametrize("verbose", (0, 1))
 def test_BinnedNLL(binned, verbose):
     mle, nx, xe = binned
@@ -128,7 +175,6 @@ def test_weighted_BinnedNLL():
     n = p * 1000
     c = BinnedNLL(n, xe, expon_cdf)
 
-    c = BinnedNLL(n, xe, expon_cdf)
     assert_equal(c.data, n)
     m1 = Minuit(c, 1)
     m1.migrad()
@@ -163,6 +209,68 @@ def test_BinnedNLL_bad_input_4():
         BinnedNLL([[1, 2, 3]], [1, 2], lambda x, a: 0)
 
 
+def test_BinnedNLL_ndof_zero():
+    c = BinnedNLL([1, 2], [0, 1, 2], lambda x, loc, scale: norm.cdf(x, loc, scale))
+    m = Minuit(c, loc=0, scale=1)
+    m.migrad()
+    assert c.ndata == m.nfit
+    assert np.isnan(m.fmin.reduced_chi2)
+
+
+def test_BinnedNLL_bad_input_5():
+    with pytest.raises(ValueError):
+        BinnedNLL([[1, 2, 3]], [[1, 2], [1, 2, 3]], lambda x, a: 0)
+
+
+def test_BinnedNLL_bad_input_6():
+    with pytest.raises(ValueError):
+        BinnedNLL(1, 2, lambda x, a: 0)
+
+
+def test_BinnedNLL_2D():
+    truth = (0.1, 0.2, 0.3, 0.4, 0.5)
+    x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T
+
+    w, xe, ye = np.histogram2d(x, y, bins=(20, 50))
+
+    def model(xy, mux, muy, sx, sy, rho):
+        return mvnorm(mux, muy, sx, sy, rho).cdf(xy.T)
+
+    cost = BinnedNLL(w, (xe, ye), model)
+    assert cost.ndata == np.prod(w.shape)
+    m = Minuit(cost, *truth)
+    m.limits["sx", "sy"] = (0, None)
+    m.limits["rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+    assert_allclose(m.values, truth, atol=0.05)
+
+    assert cost.ndata == np.prod(w.shape)
+    w2 = w.copy()
+    w2[1, 1] += 1
+    cost.n = w2
+    assert cost(*m.values) > m.fval
+
+
+def test_BinnedNLL_2D_with_zero_bins():
+    truth = (0.1, 0.2, 0.3, 0.4, 0.5)
+    x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T
+
+    w, xe, ye = np.histogram2d(x, y, bins=(50, 100), range=((-5, 5), (-5, 5)))
+    assert np.mean(w == 0) > 0.25
+
+    def model(xy, mux, muy, sx, sy, rho):
+        return mvnorm(mux, muy, sx, sy, rho).cdf(xy.T)
+
+    cost = BinnedNLL(w, (xe, ye), model)
+    m = Minuit(cost, *truth)
+    m.limits["sx", "sy"] = (0, None)
+    m.limits["rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+    assert_allclose(m.values, truth, atol=0.05)
+
+
 @pytest.mark.parametrize("verbose", (0, 1))
 def test_ExtendedBinnedNLL(binned, verbose):
     mle, nx, xe = binned
@@ -202,6 +310,52 @@ def test_ExtendedBinnedNLL_bad_input():
         ExtendedBinnedNLL([1], [1], lambda x, a: 0)
 
 
+def test_ExtendedBinnedNLL_2D():
+    truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
+    x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 1000), random_state=1).T
+
+    w, xe, ye = np.histogram2d(x, y, bins=(10, 20))
+
+    def model(xy, n, mux, muy, sx, sy, rho):
+        return n * 1000 * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))
+
+    cost = ExtendedBinnedNLL(w, (xe, ye), model)
+    assert cost.ndata == np.prod(w.shape)
+    m = Minuit(cost, *truth)
+    m.limits["n", "sx", "sy"] = (0, None)
+    m.limits["rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+    assert_allclose(m.values, truth, atol=0.1)
+
+
+def test_ExtendedBinnedNLL_3D():
+    truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7)
+    n = int(truth[0] * 10000)
+    x, y = mvnorm(*truth[1:-2]).rvs(size=n).T
+    z = norm(truth[-2], truth[-1]).rvs(size=n)
+
+    w, edges = np.histogramdd((x, y, z), bins=(5, 10, 20))
+
+    def model(xyz, n, mux, muy, sx, sy, rho, muz, sz):
+        *xy, z = xyz
+        return (
+            n
+            * 10000
+            * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))
+            * norm(muz, sz).cdf(z)
+        )
+
+    cost = ExtendedBinnedNLL(w, edges, model)
+    assert cost.ndata == np.prod(w.shape)
+    m = Minuit(cost, *truth)
+    m.limits["n", "sx", "sy", "sz"] = (0, None)
+    m.limits["rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+    assert_allclose(m.values, truth, atol=0.05)
+
+
 @pytest.mark.parametrize("loss", ["linear", "soft_l1", np.arctan])
 @pytest.mark.parametrize("verbose", (0, 1))
 def test_LeastSquares(loss, verbose):
@@ -312,13 +466,9 @@ def test_BinnedNLL_properties():
     assert_equal(c.n, [1])
     assert_equal(c.xe, [1, 2])
     c.n = [2]
-    c.xe = [2, 3]
     assert_equal(c.n, [2])
-    assert_equal(c.xe, [2, 3])
     with pytest.raises(ValueError):
         c.n = [1, 2]
-    with pytest.raises(ValueError):
-        c.xe = [1, 2, 3]
 
 
 def test_ExtendedBinnedNLL_mask():
@@ -390,6 +540,30 @@ def test_LeastSquares_properties():
         c.yerror = [1, 2]
 
 
+def test_LeastSquares_2D():
+    x = np.array([1.0, 2.0, 3.0])
+    y = np.array([4.0, 5.0, 6.0])
+    z = 1.5 * x + 0.2 * y
+    ze = 1.5
+
+    def model(xy, a, b):
+        x, y = xy
+        return a * x + b * y
+
+    c = LeastSquares((x, y), z, ze, model)
+    assert_equal(c.x, (x, y))
+    assert_equal(c.y, z)
+    assert_equal(c.yerror, ze)
+    assert_allclose(c(1.5, 0.2), 0.0)
+    assert_allclose(c(2.5, 0.2), np.sum(((z - 2.5 * x - 0.2 * y) / ze) ** 2))
+    assert_allclose(c(1.5, 1.2), np.sum(((z - 1.5 * x - 1.2 * y) / ze) ** 2))
+
+    c.y = 2 * z
+    assert_equal(c.y, 2 * z)
+    c.x = (y, x)
+    assert_equal(c.x, (y, x))
+
+
 def test_addable_cost_1():
     def model1(x, a):
         return a + x
@@ -538,15 +712,36 @@ def test_NormalConstraint_properties():
     assert_equal(nc.covariance, (1, 2))
 
 
-def test_log_poisson_part():
-    assert _log_poisson_part(0, 0) == 0
-    assert _log_poisson_part(0, 1) == 0
-    assert _log_poisson_part(1, 0) == pytest.approx(743, abs=1)
+dtypes_to_test = [np.float32]
+if hasattr(np, "float128"):  # not available on all platforms
+    dtypes_to_test.append(np.float128)
+
+
+@pytest.mark.parametrize("dtype", dtypes_to_test)
+def test_soft_l1_loss(dtype):
+    v = np.array([0], dtype=dtype)
+    assert _soft_l1_loss(v) == v
+    v[:] = 0.1
+    assert _soft_l1_loss(v) == pytest.approx(0.1, abs=0.01)
+    v[:] = 1e10
+    assert _soft_l1_loss(v) == pytest.approx(2e5, rel=0.01)
+
+
+def test_multinominal_chi2():
+    zero = np.array(0)
+    one = np.array(1)
+
+    assert multinominal_chi2(zero, zero) == 0
+    assert multinominal_chi2(zero, one) == 0
+    assert multinominal_chi2(one, zero) == pytest.approx(1487, abs=1)
     n = np.array([(0.0, 0.0)])
-    assert_allclose(_log_poisson_part(n, 0), 0)
-    assert_allclose(_log_poisson_part(n, 1), 0)
+    assert_allclose(multinominal_chi2(n, zero), 0)
+    assert_allclose(multinominal_chi2(n, one), 0)
 
 
+@pytest.mark.skipif(
+    not hasattr(np, "float128"), reason="float128 not available on all platforms"
+)
 def test_model_float128():
     def model(x, a):
         x = x.astype(np.float128)
@@ -584,12 +779,12 @@ def test_model_performance_warning():
         ExtendedUnbinnedNLL([1], lambda x, a: (1, model(x, a)))(1)
 
 
-@pytest.mark.parametrize("klass", (BinnedNLL, ExtendedBinnedNLL))
-def test_update_data_with_mask(klass):
+@pytest.mark.parametrize("cls", (BinnedNLL, ExtendedBinnedNLL))
+def test_update_data_with_mask(cls):
     xe = np.arange(0, 4)
     nx = np.diff(expon_cdf(xe, 1))
     nx[0] += 1
-    c = klass(nx.copy(), xe, expon_cdf)
+    c = cls(nx.copy(), xe, expon_cdf)
 
     c.mask = [False, True, True]
     assert c(1) == 0
@@ -603,3 +798,102 @@ def test_update_data_with_mask(klass):
     c.mask = (True, False, True)
     c.n = nx
     assert c(1) == 0
+
+
+def test_barlow_beeston_lite_chi2_jsc():
+    n = np.array([1, 2, 3])
+    r = barlow_beeston_lite_chi2_jsc(n, n, n**2)
+    assert_allclose(r, 0)
+
+    n = np.array([1, 2, 3])
+    r = barlow_beeston_lite_chi2_jsc(n, n, 10 * n**2)
+    assert_allclose(r, 0)
+
+    mu = np.array([2, 2, 3])
+    r = barlow_beeston_lite_chi2_jsc(n, mu, mu**2)
+    assert r > 0
+
+    mu = np.array([0.999, 2, 3])
+    r = barlow_beeston_lite_chi2_jsc(n, mu, mu**2)
+    assert r > 0
+
+
+@pytest.mark.parametrize("method", ("jsc", "hpd"))
+def test_BarlowBeestonLite(method):
+    n = np.array([1, 2, 3])
+    xe = np.array([0, 1, 2, 3])
+    t = np.array([[1, 1, 0], [0, 1, 3]])
+
+    c = BarlowBeestonLite(n, xe, t, method=method)
+    m = Minuit(c, 1, 1)
+    m.migrad()
+    assert m.valid
+    assert_allclose(m.fval, 0, atol=1e-4)
+    assert_allclose(m.values, [2, 4], atol=1e-2)
+
+
+def generate(rng, nmc, truth, bins, tf=1, df=1):
+    xe = np.linspace(0, 2, bins + 1)
+    b = np.diff(truncexpon(1, 0, 2).cdf(xe))
+    s = np.diff(norm(1, 0.1).cdf(xe))
+    n = b * truth[0] + s * truth[1]
+    t = b * nmc, s * nmc
+    if rng is not None:
+        n = rng.poisson(n / df) * df
+        if df != 1:
+            n = np.transpose((n, n * df))
+        t = [rng.poisson(ti / tf) * tf for ti in t]
+        if tf != 1:
+            t = [np.transpose((tj, tj * tf)) for tj in t]
+    return n, xe, np.array(t)
+
+
+@pytest.mark.parametrize("method", ("jsc", "hpd"))
+@pytest.mark.parametrize("with_mask", (False, True))
+@pytest.mark.parametrize("weighted_data", (False, True))
+def test_BarlowBeestonLite_weighted(method, with_mask, weighted_data):
+    rng = np.random.default_rng(1)
+    truth = 750, 250
+    z = []
+    rng = np.random.default_rng(1)
+    for itoy in range(100):
+        ni, xe, ti = generate(rng, 400, truth, 15, 1.5, 1.5 if weighted_data else 1)
+        c = BarlowBeestonLite(ni, xe, ti, method=method)
+        if with_mask:
+            cx = 0.5 * (xe[1:] + xe[:-1])
+            c.mask = cx != 1.5
+        m = Minuit(c, *truth)
+        m.limits = (0, None)
+        m.strategy = 0
+        for iter in range(10):
+            m.migrad(iterate=1)
+            m.hesse()
+            if m.valid and m.accurate:
+                break
+        assert m.valid
+        z.append((m.values[1] - truth[1]) / m.errors[1])
+    assert_allclose(np.mean(z), 0, atol=0.2)
+    assert_allclose(np.std(z), 1, rtol=0.15)
+
+
+@pytest.mark.parametrize("method", ("jsc", "hpd"))
+def test_BarlowBeestonLite_bad_input(method):
+    with pytest.raises(ValueError):
+        BarlowBeestonLite([1, 2], [1, 2, 3], [], method=method)
+
+    with pytest.raises(ValueError, match="do not match"):
+        BarlowBeestonLite([1, 2], [1, 2, 3], [[1, 2, 3], [1, 2, 3]], method=method)
+
+    with pytest.raises(ValueError, match="do not match"):
+        BarlowBeestonLite(
+            [1, 2],
+            [1, 2, 3],
+            [[[1, 2], [3, 4]], [[1, 2], [3, 4], [5, 6]]],
+            method=method,
+        )
+
+    with pytest.raises(ValueError, match="not understood"):
+        BarlowBeestonLite([1], [1, 2], [[1]], method="foo")
+
+    with pytest.raises(ValueError, match="number of names"):
+        BarlowBeestonLite([1], [1, 2], [[1]], name=("b", "s"))
